Turninator = LibStub("AceAddon-3.0"):NewAddon("Turninator", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Turninator")

--==== set upvalues ====--
-- TODO: find out if QuestInfoFrame can be upvalued
-- GLOBALS: LibStub, CreateFrame, QuestInfoFrame, UIParent
local Turninator, ceil, format, ipairs, pairs, print, select, strlen, strtrim, tsort, tinsert, tonumber, type, unpack = Turninator, ceil, format, ipairs, pairs, print, select, strlen, strtrim, table.sort, tinsert, tonumber, type, unpack
local AcceptQuest, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption = AcceptQuest, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption
local GetBindingByKey, GetBindingKey, GetItemInfo, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled = GetBindingByKey, GetBindingKey, GetItemInfo, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled


local DL = {
	[0] = "standard",
	"error",
	"test",
	"events",
	"config",
	"spam",
}

local dbgprint = { squelch = function(...) return nil end, 
                   print   = function(...) return Turninator:Print(...) end }
local function SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print; end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local tblCommands = { 
	DialogAction  = { strCommand = "CLICK TurninatorButton:LeftButton", strConfigMsg = "taking binding for dialog event" },
	PopUpQuest    = { strCommand = "CLICK WatchFrameAutoQuestPopUp1:LeftButton", strConfigMsg = "taking binding for quest popup" },
	UseItem       = { strCommand = nil, strConfigMsg = "taking binding for item use" },
	Interact      = { strCommand = "INTERACTTARGET", strConfigMsg = "taking binding for interaction" },
	Released      = { strCommand = nil, strConfigMsg = "releasing binding" },
}

-- Called when the addon is loaded
function Turninator:OnInitialize()
	self.db = Turninator.LoadSavedSettings()
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Turninator", Turninator.GetUIOptions, { "Turninator", "tr" } )

	for k,v in pairs(DL) do DL[v] = k end
	SetDebugLevel(self.db.global.iDebugLevel)

	-- if there's no shared command, then have the binding release to Interact with Target
	tblCommands.Released.strCommand = self.db.profile.strSharedCmd or tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}
end


-- Called when the addon is enabled
function Turninator:OnEnable()
	if (self.db.profile.fEnabled) then
		dbgprint.spam("enabling addon")

		self:RegisterEvent("GOSSIP_SHOW",     "DialogEventHandler")
		self:RegisterEvent("GOSSIP_CLOSED",   "DialogEventHandler")
		self:RegisterEvent("QUEST_GREETING",  "DialogEventHandler")
		self:RegisterEvent("QUEST_DETAIL",    "DialogEventHandler")
		self:RegisterEvent("QUEST_PROGRESS",  "DialogEventHandler")
		self:RegisterEvent("QUEST_COMPLETE",  "DialogEventHandler")
		self:RegisterEvent("QUEST_FINISHED",  "DialogEventHandler")

		self:RegisterEvent("PLAYER_TARGET_CHANGED", "SelectBinding")

		self:SecureHook("AddAutoQuestPopUp",    "AddAutoQuestPopUpHook")
		self:SecureHook("RemoveAutoQuestPopUp", "RemoveAutoQuestPopUpHook")

		self:CreateButton()

--		self:RegisterEvent("UPDATE_BINDINGS", "UpdateBindingsHandler")
	end

	self:SelectBinding("OnEnable")
end

-- Called when the addon is Disabled
function Turninator:OnDisable()
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	if (self.db.profile.fEnabled) then
		dbgprint.spam("disabling addon")
		self.tblDialogAction = {}

		self:UnregisterEvent("GOSSIP_SHOW")
		self:UnregisterEvent("GOSSIP_CLOSED")
		self:UnregisterEvent("QUEST_GREETING")
		self:UnregisterEvent("QUEST_DETAIL")
		self:UnregisterEvent("QUEST_PROGRESS")
		self:UnregisterEvent("QUEST_COMPLETE")
		self:UnregisterEvent("QUEST_FINISHED")

		self:UnregisterEvent("PLAYER_TARGET_CHANGED")

		self:UnhookAll()

--		self:UnregisterEvent("UPDATE_BINDINGS")
	end
end

function Turninator:CreateButton()
	if (not self.btn) then
		self.btn = CreateFrame("Button", "TurninatorButton", UIParent, "SecureActionButtonTemplate");
		self.btn:SetAttribute("type", "macro")
		self.btn:SetAttribute("macrotext", "/run Turninator:DialogAction()")
	end
end


function Turninator:EventHandler(...)
	dbgprint.events(...)
end

function Turninator:UpdateBindingsHandler(...)
	local strBind = self.db.profile.strSharedKey and GetBindingByKey(self.db.profile.strSharedKey) or ""
	dbgprint.events(..., strBind)
end

function Turninator:DialogEventHandler(event, ...)
	dbgprint.events(event, ...)
	self.tblDialogAction = self:GetDialogAction(event, ...)
	self:SelectBinding(event);
end

function Turninator:AddAutoQuestPopUpHook(...) -- questId, type)
	dbgprint.events("AddAutoQuestPopUpHook", ...)
	self:SelectBinding("AddAutoQuestPopUpHook")
end

function Turninator:RemoveAutoQuestPopUpHook(...) -- questId
	dbgprint.events("RemoveAutoQuestPopUpHook", ...)
	self:SelectBinding("RemoveAutoQuestPopUpHook")
end


function Turninator:Config()
	dbgprint.error("debug reporting level set to", DL[self.db.global.iDebugLevel])
	dbgprint.standard(L["addon is"], self.db.profile.fEnabled and L["enabled"] or L["disabled"])
	dbgprint.standard(L["shared cmd is"], self.db.profile.strSharedCmd or L["not set"])
	dbgprint.standard(L["shared key is"], self.db.profile.strSharedKey or L["not set"])
	if (self.db.profile.strSharedKey) then 
		dbgprint.spam("shared key currently bound to", GetBindingByKey(self.db.profile.strSharedKey))
	end
end


function Turninator:Debug(info, value)
	local strValue, index = self:GetArgs(value, 1);

	if (strValue ~= nil) then
		if (index ~= 1e9) then
			dbgprint.standard(L["ERROR: too many parameters passed"]);
		end

		local iValue = tonumber(strValue)
		if (iValue == nil) then
			dbgprint.standard(L["ERROR: no value given for debug level"]);
			return;
		elseif (DL[iValue] == nil) then
			dbgprint.standard(L["ERROR: value must be number between 0 and"], #DL);
			return;
		else
			self.db.global.iDebugLevel = iValue
			SetDebugLevel(iValue)
		end
	end

	self:Config()
end

function Turninator:Reset(silent)
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	self.db.profile.strSharedKey = nil
	self.db.profile.strSharedCmd = nil

	tblCommands.Released.strCommand = tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}

	if not silent then self:Config() end
end

function Turninator:ShareKey(info, value)
	local strKey, index = self:GetArgs(value, 1);

	if (strKey == nil) then
		dbgprint.standard(L["ERROR: no key specified"]);
		return;
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strAction = GetBindingByKey(strKey)
	if (not SetBinding(strKey, strAction)) then
		dbgprint.standard(L["ERROR: invalid key"])
		return;
	end

	self:Reset(true);  -- release old binding before setting up new one

	if (not strAction or strAction == "") then
		dbgprint.config("sharing previously unbound key");
	else
		dbgprint.config("sharing key", strKey, "with action", strAction);
		self.db.profile.strSharedCmd = strAction
		tblCommands.Released.strCommand = strAction
	end
	self.db.profile.strSharedKey = strKey

	self.strCurBind = "Released"
	self:SelectBinding("ShareKey")
	self:Config()
end

function Turninator:ShareCmd(info, value)
	local strAction, index = self:GetArgs(value, 1);

	if (strAction == nil) then
		dbgprint.standard(L["ERROR: no action specified"]);
		return;
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strKey = GetBindingKey(strAction)
	if (not strKey or strKey == "") then
		dbgprint.standard(L["ERROR: action does not exist or is unbound"]);
		return;
	end

	self:Reset(true);  -- release old binding before setting up new one

	self.db.profile.strSharedCmd = strAction
	self.db.profile.strSharedKey = strKey

	tblCommands.Released.strCommand = strAction
	self.strCurBind = "Released"
	self:SelectBinding("ShareCmd")
	self:Config()
end

function Turninator:UseItem(info, strItem)
	local strName = UnitName("target")
	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
		return
	elseif (strItem == nil) then
		dbgprint.standard(L["ERROR: no item specified"]);
		return
	end

	-- strItem could be the item link, so we need to keep the return value from GetItemInfo
	local strItemName = GetItemInfo(strItem)
	if (strItemName == nil) then
		dbgprint.standard(L["ERROR: item not found in inventory"]);
		return
	end

	dbgprint.standard("adding item", strItemName, "for use on", strName)
	self.db.global.tblUseItem[strName] = strItemName
	self:SelectBinding("UseItem")
end

function Turninator:SetNPC(strAction)
	local strName = UnitName("target")
	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
	elseif (strAction == "Interact") then
		dbgprint.standard("adding", strName, "to Interact list")
		self.db.global.tblInteract[strName] = true
		self:SelectBinding("Interact")
	elseif (strAction == "Ignore") then
		dbgprint.standard("adding", strName, "to Ignore list")
		self.db.global.tblInteract[strName] = false
		self:SelectBinding("Ignore")
	elseif (strAction == "Forget") then
		dbgprint.standard("forgetting", strName)
		self.db.global.tblInteract[strName] = nil
		self.db.global.tblUseItem[strName] = nil
		self:SelectBinding("Forget")
	end
end

function Turninator:ListNPCs()
	local ignore, interact, items = {}, {}, {}
	local cmac = 0

	for strNPCName,fInteract in pairs(self.db.global.tblInteract) do
		tinsert( (fInteract and interact or ignore), strNPCName)
	end

	if (#interact > 0) then
		dbgprint.standard(L["forcing interaction with:"])
		tsort(interact)
		for _,v in ipairs(interact) do
			print(format("%2s", v))
		end
	end

	if (#ignore > 0) then
		dbgprint.standard(L["ignoring:"])
		tsort(ignore)
		for _,v in ipairs(ignore) do
			print(format("  %s", v))
		end
	end


	for strNPCName,strItemName in pairs(self.db.global.tblUseItem) do
		tinsert( items, { name = strNPCName, item = strItemName } )
		local clen = strlen(strNPCName)
		if (cmac < clen) then cmac = clen end
	end

	if (#items) then 
		dbgprint.standard(L["npc item bindings:"])
		tsort(items, function(a,b) return (a.name < b.name) end)
		for _,v in ipairs(items) do
			print(format( format("  %%-%is%%s", cmac+3), v.name, v.item ))
		end
	end
end

function Turninator:SelectBinding(event)
	if (self.db.profile.strSharedKey == nil) then 
		if (self.db.profile.fEnabled) then dbgprint.error("ERROR: no key is set; cannot respond to", event); end
	elseif (self.db.profile.fEnabled and #self.tblDialogAction ~= 0) then
		self:TakeBinding("DialogAction", event)
	elseif (self.db.profile.fEnabled and GetNumAutoQuestPopUps() ~= 0) then
		self:TakeBinding("PopUpQuest", event)
	elseif (self.db.profile.fEnabled and self:CanUseItem()) then
		self:TakeBinding("UseItem", event)
	elseif (self.db.profile.fEnabled and self:CanInteract()) then
		self:TakeBinding("Interact", event)
	else
		self:TakeBinding("Released", event)
	end
end

function Turninator:TakeBinding(strBind, event)
	if (self.strCurBind ~= strBind) then
		if (UnitAffectingCombat("player")) then
			dbgprint.error("currently in combat; binding cannot be changed")
		else
			local bind = tblCommands[strBind]
			if (SetBinding(self.db.profile.strSharedKey, bind.strCommand)) then
				dbgprint.config(bind.strConfigMsg, event);
				self.strCurBind = strBind
			end
		end
	end
end

function Turninator:GetDialogAction(event, ...)

	if (event == "GOSSIP_SHOW") then
		local function GetActive(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 4 do
				local iActive = ceil(iOffset/4)
				local strQuest = select(iOffset, ...)
				local fComplete = select(iOffset+3, ...) ~= nil
				tbl[iActive] = { name = strQuest, complete = fComplete }
			end
			return tbl;
		end

		local function GetAvailable(...)
			local function CompareAvailableQuests(a, b)
				if (not (a.precedence == b.precedence)) then
					return a.precedence > b.precedence
				elseif (not (a.level == b.level)) then
					return a.level > b.level
				else
					return a.index < b.index
				end
			end

			local tbl = {}
			for iOffset = 1, select("#", ...), 5 do
				local index = ceil(iOffset/5)
				local strQuest, level, isTrivial, isDaily, isRepeatable = select(iOffset, ...)
				local precedence = (isRepeatable and 0) or (isTrivial and 1) or (isDaily and 2) or 3
				tinsert(tbl, { index = index, name = strQuest, level = level, precedence = precedence } )
			end
			tsort(tbl, CompareAvailableQuests)
			return tbl;
		end

		local function GetGossip(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 2 do
				local strType = select(iOffset+1, ...)
				if (strType == "gossip") then
					tbl[#tbl+1] = ceil(iOffset/2)
				end
			end
			return tbl;
		end

		-- check for quests we can turn in
		local tblActive = GetActive(GetGossipActiveQuests())
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectGossipActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumGossipAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			local tblAvailable = GetAvailable(GetGossipAvailableQuests())
			return { SelectGossipAvailableQuest, tblAvailable[1].index }
		end

		-- if the npc has chatty gossip options try the first one
		if (GetNumGossipOptions() > 0) then
			local tblGossip = GetGossip(GetGossipOptions())
			-- PrintTable(tblGossip)
			dbgprint.spam(#tblGossip, "gossip text options for this npc")
			if (tblGossip[1] ~= nil) then 
				return { SelectGossipOption, tblGossip[1] }
			end
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseGossip }

	elseif (event == "QUEST_GREETING") then
		local function GetActive()
			local tbl = {}
			for iActive = 1, GetNumActiveQuests() do
				local strQuest, fComplete = GetActiveTitle(iActive)
				tbl[iActive] = { name = strQuest, complete = (fComplete ~= nil) }
			end
			return tbl;
		end

		local tblActive = GetActive()
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			return { SelectAvailableQuest, 1 }
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseQuest }

	elseif (event == "QUEST_DETAIL") then
		return { AcceptQuest }

	elseif (event == "QUEST_PROGRESS") then
		if (IsQuestCompletable()) then
			return { CompleteQuest }
		else
			return { DeclineQuest }
		end
		-- todo - can't complete, back up?

	elseif (event == "QUEST_COMPLETE") then
		local function CheckRewardSelection()
			if (QuestInfoFrame.itemChoice ~= 0) then
				GetQuestReward(QuestInfoFrame.itemChoice)
			else
				QuestChooseRewardError()
			end 
		end

		if (GetNumQuestChoices() == 0) then
			return { GetQuestReward }
		else
			return { CheckRewardSelection }
		end

	elseif (event == "GOSSIP_CLOSED") then
		return {}

	elseif (event == "QUEST_FINISHED") then
		return {}
	end

	return {}
end

function Turninator:DialogAction()
	if (not self.tblDialogAction or type(self.tblDialogAction) ~= "table") then
		dbgprint.error("ERROR - tblDialogAction is of type", type(self.tblDialogAction));
	elseif (not self.tblDialogAction[1] or type(self.tblDialogAction[1]) ~= "function") then
		dbgprint.error("ERROR - tblDialogAction does not contain function");
	else
		self.tblDialogAction[1](unpack(self.tblDialogAction, 2));
	end
end

function Turninator:CanUseItem()
	local name = UnitName("target")
	if (not name or UnitPlayerControlled("target")) then
		return false
	end

	local item = self.db.global.tblUseItem[name]
	-- dbgprint.test(item)
	if (not item or not IsUsableItem(item)) then
		return false
	end

	-- set the binding up in tblCommands
	tblCommands.UseItem.strCommand = format("ITEM %s", item)
	-- print(tblCommands.UseItem.strCommand)
	return true
end

function Turninator:CanInteract()
	local name = UnitName("target")
	if (not name or UnitPlayerControlled("target")) then
		return false
	end

	local interact = self.db.global.tblInteract[name]
	if (interact ~= nil) then
		return interact
	end

	return (UnitIsFriend("player", "target") and not UnitIsCorpse("target") 
		and not UnitIsDead("target") and not UnitCanAttack("player", "target"))
end

