local Turninator = Turninator
local L = LibStub("AceLocale-3.0"):GetLocale("Turninator")


function Turninator.LoadSavedSettings()
	-- saved settings and default values
	local defaults = {
		global = {
			iDebugLevel = 0,
			tblInteract = {},   -- table { key(string) name of npc, value(boolean)
			tblUseItem  = {},   -- table { key(string) name of npc, value(string) name of item to use
		},
		profile = {
			iDataFormat = 0,
			fEnabled = true,
		},
	}

	return LibStub("AceDB-3.0"):New("TurninatorDB", defaults)
end


--=========================================================================--
-- Options UI definition
--=========================================================================--

function Turninator.GetUIOptions()
	if (not Turninator.options) then
		Turninator.options = {
			name = "Turninator",
			handler = Turninator,
			type = 'group',
--			childGroups = "tab",
			args = {
				Debug = {
					hidden = true,
					order = 0,
					type = "input",
					name = "debug",
					desc = "Sets verbosity of debug output",
					get = function(info) return "" end,
					set = "Debug",
				},
				[ L["on"] ] = {
					hidden = guiHidden,
					order = 1,
					type = "execute",
					name = L["on"],
					desc = L["Enables the addon"],
					func = function(info) 
						if (not Turninator.db.profile.fEnabled) then
							Turninator.db.profile.fEnabled = true
							Turninator:OnEnable()
						end
						Turninator:Config()
					end,
				},
				[ L["off"] ] = {
					hidden = guiHidden,
					order = 2,
					type = "execute",
					name = L["off"],
					desc = L["Disables the addon"],
					func = function(info) 
						if (Turninator.db.profile.fEnabled) then 
							Turninator:OnDisable()
							Turninator.db.profile.fEnabled = false
						end
						Turninator:Config()
					end,
				},
				[ L["config"] ] = {
					hidden = guiHidden,
					order = 3,
					type = "execute",
					name = L["config"],
					desc = L["display current configuration"],
					func = function(info) Turninator:Config() end,
				},
				[ L["ShareKey"] ] = {
					hidden = guiHidden,
					order = 4,
					type = "input",
					name = L["ShareKey"],
					desc = L["Sets the keybind to be used/shared by the addon"],
					get = function(info) return "" end,
					set = "ShareKey",
				},
				[ L["ShareCmd"] ] = {
					hidden = true,
					order = 5,
					type = "input",
					name = L["ShareCmd"],
					desc = L["Sets the addon to use/share the keybind for that command"],
					get = function(info) return "" end,
					set = "ShareCmd",
				},
				[ L["Reset"] ] = {
					hidden = guiHidden,
					order = 6,
					type = "execute",
					name = L["Reset"],
					desc = L["Decouples the addon from the shared keybind"],
					func = function(info) Turninator:Reset() end,
				},
				[ L["UseItem"] ] = {
					hidden = guiHidden,
					order = 7,
					type = "input",
					name = L["UseItem"],
					desc = L["Set binding to always use named item on targeted npc"],
					get = function(info) return "" end,
					set = "UseItem",
				},
				[ L["Interact"] ] = {
					hidden = guiHidden,
					order = 8,
					type = "execute",
					name = L["Interact"],
					desc = L["Treat targeted npc as interactable from now on"],
					func = function(info) Turninator:SetNPC("Interact") end,
				},
				[ L["Ignore"] ] = {
					hidden = guiHidden,
					order = 9,
					type = "execute",
					name = L["Ignore"],
					desc = L["Treat targeted npc as not being interactable from now on"],
					func = function(info) Turninator:SetNPC("Ignore") end,
				},
				[ L["Forget"] ] = {
					hidden = guiHidden,
					order = 10,
					type = "execute",
					name = L["Forget"],
					desc = L["Removes targeted npc from UseItem, Interact and/or Ignore lists"],
					func = function(info) Turninator:SetNPC("Forget") end,
				},
				[ L["List"] ] = {
					hidden = guiHidden,
					order = 11,
					type = "execute",
					name = L["List"],
					desc = L["Prints out the current contents of the UseItem, Interact and/or Ignore lists"],
					func = function(info) Turninator:ListNPCs() end,
				},
			},
		}
	end

	return Turninator.options;
end
